import json


class Mythic:
    def __init__(
        self,
        username: str = None,
        password: str = None,
        apitoken: str = None,
        access_token: str = None,
        refresh_token: str = None,
        server_ip: str = None,
        ssl: bool = False,
        server_port: int = None,
        server_api_version: float = 1.4,
        global_timeout: int = None,
        schema: str = None,
    ):
        self.username = username
        self.password = password
        self.apitoken = apitoken
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.server_ip = server_ip
        self.server_port = server_port
        self.server_api_version = server_api_version
        self.ssl = ssl
        self.http = "http://" if not ssl else "https://"
        self.ws = "ws://" if not ssl else "wss://"
        self.global_timeout = global_timeout if global_timeout is not None else -1
        self.scripting_version = 3
        self.current_operation_id = 0
        self.operator_id = None
        self.operator = None
        self.schema = schema

    def __str__(self):
        return json.dumps(
            {
                "username": self.username,
                "password": self.password,
                "apitoken": self.apitoken,
                "access_token": self.access_token,
                "refresh_token": self.refresh_token,
                "server_ip": self.server_ip,
                "server_port": self.server_port,
                "ssl": self.ssl,
                "current_operation_id": self.current_operation_id,
                "operator": self.operator,
                "operator_id": self.operator_id,
            },
            indent=4,
        )


class MythicStatus:
    Error = "error"
    Completed = "completed"
    Processed = "processed"
    Processing = "processing"
    Preprocessing = "preprocessing"
    Delegating = "delegating"
    Submitted = "submitted"

    def __init__(self, status: str):
        self.status = status

    def __str__(self):
        return self.status

    def __eq__(self, obj):
        # override == operator
        if isinstance(obj, str):
            return self.status == obj
        elif isinstance(obj, MythicStatus):
            return self.status == obj.status
        else:
            return False

    def __ge__(self, obj):
        # override >= operator
        # self.status is lefthand side, obj is right hand side
        target_obj = obj.status.lower() if isinstance(obj, MythicStatus) else obj.lower()
        if "delegating" in target_obj:
            target_obj = "delegating"
        elif "error" in target_obj:
            target_obj = "error"
        self_obj = self.status.lower()
        if "delegating" in self_obj:
            self_obj = "delegating"
        if "error" in self_obj:
            # all errors are >= other steps in the status line
            return True
        elif MythicStatus.Completed == self_obj:
            # completed status is >= all other steps in line
            return True
        enum_mapping = {
            "preprocessing": 0,
            "submitted": 1,
            "delegating": 2,
            "processing": 3,
            "processed": 4,
            "completed": 5,
            "error": 6,
        }
        if target_obj not in enum_mapping:
            raise ValueError(f"Can't compare status of type: {target_obj}")
        elif self_obj not in enum_mapping:
            raise ValueError(f"Can't compare status of type: {self_obj}")
        else:
            return enum_mapping[self_obj] >= enum_mapping[target_obj]



import asyncio
import logging
import ssl
from os import path
from time import time
from typing import AsyncGenerator, TypeVar

import aiohttp
from gql import Client, gql
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.websockets import WebsocketsTransport
from graphql import print_schema

from . import mythic_classes

T = TypeVar("T")


async def timeout_generator(
    it: AsyncGenerator[T, None], timeout: float = None
) -> AsyncGenerator[T, None]:
    start = time()
    should_timeout = timeout is not None and timeout > 0
    while True:
        # we don't want to wait for timeout each time, that's an overall timeout
        if should_timeout:
            # the new timeout should be the total amount of time remaining
            new_timeout = timeout - (time() - start)
            if new_timeout <= 0:
                # we already hit our timeout and should return
                break
        else:
            new_timeout = None
        try:
            task = asyncio.create_task(it.__anext__())
            yield await asyncio.wait_for(task, new_timeout)
        except asyncio.TimeoutError:
            logging.info("Timeout reached in timeout_generator")
            task.cancel()
            return
        except StopAsyncIteration:
            return
        except Exception as e:
            raise e


def get_headers(mythic: mythic_classes.Mythic) -> dict:
    headers = {}
    if mythic.apitoken is not None:
        headers["apitoken"] = mythic.apitoken
    elif mythic.access_token is not None:
        headers["Authorization"] = f"Bearer {mythic.access_token}"
    return headers


async def get_http_transport(mythic: mythic_classes.Mythic) -> AIOHTTPTransport:
    transport = AIOHTTPTransport(
        url=f"{mythic.http}{mythic.server_ip}:{mythic.server_port}/graphql/",
        headers=get_headers(mythic),
    )
    return transport


async def get_ws_transport(mythic: mythic_classes.Mythic) -> WebsocketsTransport:
    ssl_context = ssl.create_default_context()
    ssl_context.check_hostname = False
    ssl_context.verify_mode = ssl.CERT_NONE
    websocket_transport = WebsocketsTransport(
        url=f"{mythic.ws}{mythic.server_ip}:{mythic.server_port}/graphql/",
        headers=get_headers(mythic),
        ssl=ssl_context,
    )
    return websocket_transport


async def http_post(mythic: mythic_classes.Mythic, data: dict, url: str) -> dict:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url, json=data, headers=get_headers(mythic), ssl=False
            ) as resp:
                return await resp.json()
    except Exception as e:
        raise e


async def http_post_form(
    mythic: mythic_classes.Mythic, data: aiohttp.FormData, url: str
) -> dict:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url, data=data, headers=get_headers(mythic), ssl=False
            ) as resp:
                return await resp.json()
    except Exception as e:
        raise e


async def http_get_dictionary(mythic: mythic_classes.Mythic, url: str) -> dict:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=get_headers(mythic), ssl=False) as resp:
                return await resp.json()
    except Exception as e:
        raise e


async def http_get(mythic: mythic_classes.Mythic, url: str) -> bytes:
    try:
        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=get_headers(mythic), ssl=False) as resp:
                return await resp.content.read()
    except Exception as e:
        raise e


async def graphql_post(
    mythic: mythic_classes.Mythic,
    gql_query: gql = None,
    query: str = None,
    variables: dict = None,
):
    try:
        query_data = gql(query) if query is not None else gql_query
        if query_data is None:
            raise Exception("No data or gql_data passed into graphql_post function")
        async with Client(
            transport=await get_http_transport(mythic=mythic),
            fetch_schema_from_transport=False,
            schema=mythic.schema,
        ) as session:
            result = await session.execute(query_data, variable_values=variables)
            return result
    except Exception as e:
        raise e


async def graphql_subscription(
    mythic: mythic_classes.Mythic,
    gql_query: gql = None,
    query: str = None,
    variables: dict = None,
    timeout: int = None,
) -> AsyncGenerator:
    try:
        query_data = gql(query) if query is not None else gql_query
        local_timeout = timeout
        if local_timeout is None:
            local_timeout = mythic.global_timeout

        if query_data is None:
            raise Exception(
                "No data or gql_data passed into graphql_subscription function"
            )

        async with Client(
            transport=await get_ws_transport(mythic=mythic),
            fetch_schema_from_transport=False,
        ) as session:
            logging.debug(f"Started subscription for {query}")
            async for result in timeout_generator(
                it=session.subscribe(query_data, variable_values=variables),
                timeout=local_timeout,
            ):
                yield result
    except Exception as e:
        print("got exception in graphql_subscription")
        raise e


async def fetch_graphql_schema(mythic: mythic_classes.Mythic):
    try:
        async with Client(
            transport=await get_http_transport(mythic=mythic),
            fetch_schema_from_transport=True,
            schema=mythic.schema,
        ) as session:
            schema = print_schema(session.client.schema)
            return schema
    except Exception as e:
        raise e


async def load_mythic_schema(mythic: mythic_classes.Mythic) -> bool:
    if path.exists("mythic_schema.graphql"):
        try:
            with open("mythic_schema.graphql", "r") as f:
                schema = f.read()
                mythic.schema = schema
            return True
        except Exception as e:
            logging.error(
                f"[-] Found mythic_schema.graphql locally, but failed to read it:\n{str(e)}"
            )
            logging.error(
                "[-] Unable to verify GraphQL queries syntactically before executing"
            )
            return False
    else:
        try:
            schema = await fetch_graphql_schema(mythic)
        except Exception as e:
            logging.error(f"[-] Failed to contact Mythic and fetch schema:\n{str(e)}")
            return False
        try:
            mythic.schema = schema
            with open("mythic_schema.graphql", "w") as f:
                f.write(schema)
            return True
        except Exception as e:
            logging.error(f"[-] Failed to save Mythic schema to disk:\n{str(e)}")
            logging.error(
                "[-] Unable to verify GraphQL queries syntactically before executing"
            )
            return False
